#include "Hero.h"
#include "Bullet.h"
#include "Saucer.h"
#include "LogManager.h"
#include "InputManager.h"
#include "ResourceManager.h"
#include "WorldManager.h"
#include "EventStep.h"
#include "GameManager.h"
#include "EventNuke.h"
#include "EventScore.h"
#include "EventLife.h"
#include "EventMult.h"
#include "EventStreakEnd.h"
#include "GameOver.h"
#include "Explosion.h"
#include "GraphicsManager.h"
#include "Display.h"

#include <iostream>
#include <sstream>
#include <string>

Hero::Hero() {
  // link to "ship" sprite
  ResourceManager &rm = ResourceManager::getInstance();
  LogManager &lm = LogManager::getInstance();
  Sprite *p_temp_sprite;
  p_temp_sprite = rm.getSprite("ship");
  if (!p_temp_sprite) {
    lm.writeLog("Hero::Hero(): Warning! Sprite '%s' not found", "ship");
    return;
  }
  setSprite(p_temp_sprite, true);
  setSpriteSlowdown(3);		  // third speed animation

  // player controls hero, so register with keyboard
  InputManager &im = InputManager::getInstance();
  im.registerInterest(this, KEYBOARD_EVENT);

  // register for all necessary events
  WorldManager &world_manager = WorldManager::getInstance();
  world_manager.registerInterest(this, STEP_EVENT);
  world_manager.registerInterest(this, SCORE_EVENT);
  world_manager.registerInterest(this, LIFE_EVENT);
  world_manager.registerInterest(this, MULT_EVENT);
  world_manager.registerInterest(this, STREAK_END_EVENT);

  setType("Hero");

  // set starting location
  pos.setX(7);
  pos.setY(world_manager.getBoundary().getVertical()/2);

  fire_slowdown = 5;
  fire_countdown = fire_slowdown;

  saucer_countdown = 500;
  saucer_consec = 0;
  nuke_count = 1;

  score_next_nuke = 100000;

  max_saucer = 100;

  score = 0;
  multiplier = 1.0;
  lives = 3;
}

Hero::~Hero() {
  InputManager &im = InputManager::getInstance();
  im.unregisterInterest(this, KEYBOARD_EVENT);
  WorldManager &world_manager = WorldManager::getInstance();
  world_manager.unregisterInterest(this, STEP_EVENT);
  world_manager.unregisterInterest(this, SCORE_EVENT);
  world_manager.unregisterInterest(this, LIFE_EVENT);
  world_manager.unregisterInterest(this, MULT_EVENT);
  world_manager.unregisterInterest(this, STREAK_END_EVENT);

  string temp = "gameover";
  GameOver *p_go = new GameOver(temp);
}

int Hero::eventHandler(Event *e) {

  if (e->getType() == KEYBOARD_EVENT) {
    EventKeyboard *p_keyboard_event = static_cast <EventKeyboard *> (e);
    kbd(p_keyboard_event);
    return 1;
  }

  if (e->getType() == STEP_EVENT) {
    step();
    hud_string();
    return 1;
  }
  // If a score event, add 100 points to the total.
  // Add another nuke if the score is high enough
  if (e->getType() == SCORE_EVENT) {
     add_score(100);
     if(score > score_next_nuke && score < 1000000){
       nuke_count++;
       score_next_nuke += 200000;
     }
     // Since only way to add score is when a saucer is shot, increase
     // the consecutive number of saucers hit.
     add_consec(1);
     return 1;
  }
  // If a life event, subtract a life. If no more lives, mark self for deletion.  Also, half the score multiplier.
  if (e->getType() == LIFE_EVENT) {
     WorldManager &world_manager = WorldManager::getInstance();
     EventMult *mult = new EventMult( -(multiplier / 2) );
     world_manager.onEvent(mult);
     add_life(-1);
     if(lives <= 0){
	world_manager.markForDelete(this);
     }
     return 1;
  }
  // If a multiplier event, add the event's multipler to the total.
  if (e->getType() == MULT_EVENT) {
     add_multiplier( static_cast<EventMult*>(e)->get_mult() );
     return 1;
  }
  // If the streak has ended, lower the streak to 0.
  if (e->getType() == STREAK_END_EVENT) {
     saucer_consec = 0;
     return 1;
  }
  // if we get here, we have ignored this event
  return 0;
}

// call move (or do nothing) according to key pressed
void Hero::kbd(EventKeyboard *p_keyboard_event) {

  switch(p_keyboard_event->getKey()) {
  case KEY_UP:			// up arrow
    move(0, -1);
    break;
  case KEY_DOWN:		// down arrow
    move(0, +1);
    break;
  case KEY_LEFT:		// left arrow
    move(-3, 0);
    break;
  case KEY_RIGHT:		// right arrow
    move(3, 0);
    break;
  case 32:			// fire
    fire();
    break;
  case 13:			// nuke!
    nuke();
    break;
  };
  return;
}

// move up or down
void Hero::move(int dx, int dy) {
  WorldManager &world_manager = WorldManager::getInstance();
  Position new_pos(pos.getX() + dx, pos.getY() + dy);

  // if stays on screen, allow move
  if (	(new_pos.getX() > 0) &&
	(new_pos.getY() > 0) &&
	(new_pos.getX() < world_manager.getBoundary().getHorizontal()) &&
      	(new_pos.getY() < world_manager.getBoundary().getVertical()) )
    world_manager.moveObj(this, new_pos);
}

// fire a bullet
void Hero::fire() {
  if (fire_countdown > 0)
    return;
  fire_countdown = fire_slowdown;
  new Bullet(pos);
}

// decrease fire restriction
// saucers spawn faster when the score is higher
// Only 100 saucers can ever be spawned here to reduce lag
void Hero::step() {
  fire_countdown--;
  if (fire_countdown < 0)
    fire_countdown = 0;
  saucer_countdown--;
  if (saucer_countdown < 0 && max_saucer > 0){
    new Saucer;
    max_saucer--;
    double temp = score / 1000;
    if(score > 0)
    	saucer_countdown = 500 - (score / 1000);
    else
        saucer_countdown = 100;
  }
}

// send nuke event to all objects
void Hero::nuke() {

  // check if nukes left
  if (!nuke_count)
    return;
  nuke_count--;

  // create a "nuke" event and send to interested
  WorldManager &world_manager = WorldManager::getInstance();
  EventNuke nuke;
  world_manager.onEvent(&nuke);
}

// Construct the hud string
// Seemingly very inefficient, and I'm sure there is a better way to
// do this, but it was stated we shouldn't worry about efficiency much.
void Hero::hud_string(){
  ostringstream hud(ostringstream::out);
  hud << "Score: ";
  hud << score;
  display_string(hud.str(), Position(0, 0));

  ostringstream hud2(ostringstream::out);
  hud2 << "Lives: ";
  if(lives > 5){
    hud2 << lives;
    hud2 << "X";
  } else {
    for(int i = 0; i < lives; i++){
       hud2 << "X";
    }
  }
  display_string(hud2.str(), Position(15, 0));

  ostringstream hud3(ostringstream::out);
  hud3 << "Nukes: ";
  if(nuke_count > 5){
    hud3 << nuke_count;
    hud3 << "X";
  } else {
    for(int i = 0; i < nuke_count; i++){
       hud3 << "X";
    }
  }
  display_string(hud3.str(), Position(28, 0));

  ostringstream hud4(ostringstream::out);
  hud4 << "Multiplier: ";
  hud4 << multiplier;
  display_string(hud4.str(), Position(40, 0));

  ostringstream hud5(ostringstream::out);
  hud5 << "Streak: ";
  hud5 << saucer_consec;
  hud5 << " / 100";
  display_string(hud5.str(), Position(58, 0));
}

// Basically setters; called by other files to adjust the lives, score, etc.
inline void Hero::add_score(int scr){
  score += static_cast<int>(scr * multiplier);
}

inline void Hero::add_life(int life){
  lives += life;
}

void Hero::add_multiplier(double mult){
  multiplier += mult;
  if(multiplier < 0)
    multiplier = 0;
}

void Hero::add_consec(int consec){
  saucer_consec++;
  if(saucer_consec >= 100){
     add_life(1);
     saucer_consec = 0;
  }
}
  
